@@ -11,6 +11,7 @@ Tests for the console kernel.
 
 # Standard library imports
 import ast
+import asyncio
 import os
 import os.path as osp
 from textwrap import dedent
@@ -19,44 +20,50 @@ import time
 from subprocess import Popen, PIPE
 import sys
 import inspect
+import uuid
 from collections import namedtuple
 
 # Test imports
-import ipykernel
-import pytest
 from flaky import flaky
+from IPython.core import release as ipython_release
 from jupyter_core import paths
 from jupyter_client import BlockingKernelClient
 import numpy as np
+import pytest
 
 # Local imports
-from spyder_kernels.py3compat import PY2, PY3, to_text_string
-from spyder_kernels.utils.iofuncs import iofunctions
-from spyder_kernels.utils.mpl import MPL_BACKENDS_FROM_SPYDER
-from spyder_kernels.utils.test_utils import get_kernel, get_log_text
+from spyder_kernels.comms.commbase import CommBase
 from spyder_kernels.customize.spyderpdb import SpyderPdb
-
-# For ipykernel 6
-try:
-    import asyncio
-except ImportError:
-    pass
+from spyder_kernels.utils.iofuncs import iofunctions
+from spyder_kernels.utils.pythonenv import PythonEnvType
+from spyder_kernels.utils.test_utils import get_kernel, get_log_text
 
 # =============================================================================
-# Constants
+# Constants and utility functions
 # =============================================================================
 FILES_PATH = os.path.dirname(os.path.realpath(__file__))
-IPYKERNEL_6 = ipykernel.__version__[0] >= '6'
 TIMEOUT = 15
 SETUP_TIMEOUT = 60
 
+# We declare this constant immediately before the test, as determining
+# that TURTLE_ACTIVE is True will briefly pop up a window, similar to the
+# windows that will pop up during the test itself.
+TURTLE_ACTIVE = False
+try:
+    import turtle
+    turtle.Screen()
+    turtle.bye()
+    TURTLE_ACTIVE = True
+except:
+    pass
+
 
 @contextmanager
 def setup_kernel(cmd):
     """start an embedded kernel in a subprocess, and wait for it to be ready
 
     This function was taken from the ipykernel project.
-    We plan to remove it when dropping support for python 2.
+    We plan to remove it.
 
     Yields
     -------
@@ -77,8 +84,6 @@ def setup_kernel(cmd):
 
         if kernel.poll() is not None:
             o,e = kernel.communicate()
-            if not PY3 and isinstance(e, bytes):
-                e = e.decode()
             raise IOError("Kernel failed to start:\n%s" % e)
 
         if not os.path.exists(connection_file):
@@ -87,7 +92,16 @@ def setup_kernel(cmd):
             raise IOError("Connection file %r never arrived" % connection_file)
 
         client = BlockingKernelClient(connection_file=connection_file)
-        client.load_connection_file()
+        tic = time.time()
+        while True:
+            try:
+                client.load_connection_file()
+                break
+            except ValueError:
+                # The file is not written yet
+                if time.time() > tic + SETUP_TIMEOUT:
+                    # Give up after 5s
+                    raise IOError("Kernel failed to write connection file")
         client.start_channels()
         client.wait_for_ready()
         try:
@@ -95,8 +109,87 @@ def setup_kernel(cmd):
         finally:
             client.stop_channels()
     finally:
-        if not PY2:
-            kernel.terminate()
+        kernel.terminate()
+
+
+class Comm():
+    """
+    Comm base class, copied from qtconsole without the qt stuff
+    """
+
+    def __init__(self, target_name, kernel_client,
+                 msg_callback=None, close_callback=None):
+        """
+        Create a new comm. Must call open to use.
+        """
+        self.target_name = target_name
+        self.kernel_client = kernel_client
+        self.comm_id = uuid.uuid1().hex
+        self._msg_callback = msg_callback
+        self._close_callback = close_callback
+        self._send_channel = self.kernel_client.shell_channel
+
+    def _send_msg(self, msg_type, content, data, metadata, buffers):
+        """
+        Send a message on the shell channel.
+        """
+        if data is None:
+            data = {}
+        if content is None:
+            content = {}
+        content['comm_id'] = self.comm_id
+        content['data'] = data
+        msg = self.kernel_client.session.msg(
+            msg_type, content, metadata=metadata)
+        if buffers:
+            msg['buffers'] = buffers
+        return self._send_channel.send(msg)
+
+    # methods for sending messages
+    def open(self, data=None, metadata=None, buffers=None):
+        """Open the kernel-side version of this comm"""
+        return self._send_msg(
+            'comm_open', {'target_name': self.target_name},
+            data, metadata, buffers)
+
+    def send(self, data=None, metadata=None, buffers=None):
+        """Send a message to the kernel-side version of this comm"""
+        return self._send_msg(
+            'comm_msg', {}, data, metadata, buffers)
+
+    def close(self, data=None, metadata=None, buffers=None):
+        """Close the kernel-side version of this comm"""
+        return self._send_msg(
+            'comm_close', {}, data, metadata, buffers)
+
+    def on_msg(self, callback):
+        """Register a callback for comm_msg
+
+        Will be called with the `data` of any comm_msg messages.
+
+        Call `on_msg(None)` to disable an existing callback.
+        """
+        self._msg_callback = callback
+
+    def on_close(self, callback):
+        """Register a callback for comm_close
+
+        Will be called with the `data` of the close message.
+
+        Call `on_close(None)` to disable an existing callback.
+        """
+        self._close_callback = callback
+
+    # methods for handling incoming messages
+    def handle_msg(self, msg):
+        """Handle a comm_msg message"""
+        if self._msg_callback:
+            return self._msg_callback(msg)
+
+    def handle_close(self, msg):
+        """Handle a comm_close message"""
+        if self._close_callback:
+            return self._close_callback(msg)
 
 
 # =============================================================================
@@ -135,15 +228,13 @@ def kernel(request):
             'False_',
             'True_'
         ],
-        'minmax': False
+        'minmax': False,
+        'filter_on':True
     }
 
     # Teardown
     def reset_kernel():
-        if IPYKERNEL_6:
-            asyncio.run(kernel.do_execute('reset -f', True))
-        else:
-            kernel.do_execute('reset -f', True)
+        asyncio.run(kernel.do_execute('reset -f', True))
     request.addfinalizer(reset_kernel)
 
     return kernel
@@ -188,10 +279,7 @@ def test_get_namespace_view(kernel):
     """
     Test the namespace view of the kernel.
     """
-    if IPYKERNEL_6:
-        execute = asyncio.run(kernel.do_execute('a = 1', True))
-    else:
-        execute = kernel.do_execute('a = 1', True)
+    asyncio.run(kernel.do_execute('a = 1', True))
 
     nsview = repr(kernel.get_namespace_view())
     assert "'a':" in nsview
@@ -199,21 +287,39 @@ def test_get_namespace_view(kernel):
     assert "'size': 1" in nsview
     assert "'view': '1'" in nsview
     assert "'numpy_type': 'Unknown'" in nsview
+    assert "'python_type': 'int'" in nsview
 
-    if PY3:
-        assert "'python_type': 'int'" in nsview
+
+@pytest.mark.parametrize("filter_on", [True, False])
+def test_get_namespace_view_filter_on(kernel, filter_on):
+    """
+    Test the namespace view of the kernel with filters on and off.
+    """
+    asyncio.run(kernel.do_execute('a = 1', True))
+    asyncio.run(kernel.do_execute('TestFilterOff = 1', True))
+
+    settings = kernel.namespace_view_settings
+    settings['filter_on'] = filter_on
+    settings['exclude_capitalized'] = True
+    nsview = kernel.get_namespace_view()
+
+    if not filter_on:
+        assert 'a' in nsview
+        assert 'TestFilterOff' in nsview
     else:
-        assert "'python_type': u'int'" in nsview
+        assert 'TestFilterOff' not in nsview
+        assert 'a' in nsview
+
+    # Restore settings for other tests
+    settings['filter_on'] = True
+    settings['exclude_capitalized'] = False
 
 
 def test_get_var_properties(kernel):
     """
     Test the properties fo the variables in the namespace.
     """
-    if IPYKERNEL_6:
-        asyncio.run(kernel.do_execute('a = 1', True))
-    else:
-        kernel.do_execute('a = 1', True)
+    asyncio.run(kernel.do_execute('a = 1', True))
 
     var_properties = repr(kernel.get_var_properties())
     assert "'a'" in var_properties
@@ -231,10 +337,7 @@ def test_get_var_properties(kernel):
 def test_get_value(kernel):
     """Test getting the value of a variable."""
     name = 'a'
-    if IPYKERNEL_6:
-        asyncio.run(kernel.do_execute("a = 124", True))
-    else:
-        kernel.do_execute("a = 124", True)
+    asyncio.run(kernel.do_execute("a = 124", True))
 
     # Check data type send
     assert kernel.get_value(name) == 124
@@ -243,10 +346,7 @@ def test_get_value(kernel):
 def test_set_value(kernel):
     """Test setting the value of a variable."""
     name = 'a'
-    if IPYKERNEL_6:
-         asyncio.run(kernel.do_execute('a = 0', True))
-    else:
-        kernel.do_execute('a = 0', True)
+    asyncio.run(kernel.do_execute('a = 0', True))
     value = 10
     kernel.set_value(name, value)
     log_text = get_log_text(kernel)
@@ -260,10 +360,7 @@ def test_set_value(kernel):
 def test_remove_value(kernel):
     """Test the removal of a variable."""
     name = 'a'
-    if IPYKERNEL_6:
-        asyncio.run(kernel.do_execute('a = 1', True))
-    else:
-        kernel.do_execute('a = 1', True)
+    asyncio.run(kernel.do_execute('a = 1', True))
 
     var_properties = repr(kernel.get_var_properties())
     assert "'a'" in var_properties
@@ -285,10 +382,7 @@ def test_copy_value(kernel):
     """Test the copy of a variable."""
     orig_name = 'a'
     new_name = 'b'
-    if IPYKERNEL_6:
-         asyncio.run(kernel.do_execute('a = 1', True))
-    else:
-        kernel.do_execute('a = 1', True)
+    asyncio.run(kernel.do_execute('a = 1', True))
 
     var_properties = repr(kernel.get_var_properties())
     assert "'a'" in var_properties
@@ -324,11 +418,7 @@ def test_load_npz_data(kernel, load):
     namespace_file = osp.join(FILES_PATH, 'load_data.npz')
     extention = '.npz'
     overwrite, execute, variables = load
-
-    if IPYKERNEL_6:
-        asyncio.run(kernel.do_execute(execute, True))
-    else:
-        kernel.do_execute(execute, True)
+    asyncio.run(kernel.do_execute(execute, True))
 
     kernel.load_data(namespace_file, extention, overwrite=overwrite)
     for var, value in variables.items():
@@ -356,11 +446,7 @@ def test_load_data(kernel):
 def test_save_namespace(kernel):
     """Test saving the namespace into filename."""
     namespace_file = osp.join(FILES_PATH, 'save_data.spydata')
-
-    if IPYKERNEL_6:
-        asyncio.run(kernel.do_execute('b = 1', True))
-    else:
-        kernel.do_execute('b = 1', True)
+    asyncio.run(kernel.do_execute('b = 1', True))
 
     kernel.save_namespace(namespace_file)
     assert osp.isfile(namespace_file)
@@ -388,7 +474,7 @@ def test_get_doc(kernel):
 def test_get_source(kernel):
     """Test to get object source."""
     objtxt = 'help'
-    assert 'class _Helper(object):' in kernel.get_source(objtxt)
+    assert 'class _Helper' in kernel.get_source(objtxt)
 
 
 # --- Other stuff
@@ -404,11 +490,7 @@ libc.printf(('Hello from C\\n').encode('utf8'))
 
     # With Wurlitzer we have the expected output
     kernel._load_wurlitzer()
-
-    if IPYKERNEL_6:
-        asyncio.run(kernel.do_execute(code, True))
-    else:
-        kernel.do_execute(code, True)
+    asyncio.run(kernel.do_execute(code, True))
     captured = capsys.readouterr()
     assert captured.out == "Hello from C\n"
 
@@ -423,11 +505,9 @@ def test_cwd_in_sys_path():
     cmd = "from spyder_kernels.console import start; start.main()"
 
     with setup_kernel(cmd) as client:
-        msg_id = client.execute("import sys; sys_path = sys.path",
-                                user_expressions={'output':'sys_path'})
-        reply = client.get_shell_msg(timeout=TIMEOUT)
-        while 'user_expressions' not in reply['content']:
-            reply = client.get_shell_msg(timeout=TIMEOUT)
+        reply = client.execute_interactive(
+            "import sys; sys_path = sys.path",
+            user_expressions={'output':'sys_path'}, timeout=TIMEOUT)
 
         # Transform value obtained through user_expressions
         user_expressions = reply['content']['user_expressions']
@@ -439,19 +519,16 @@ def test_cwd_in_sys_path():
 
 
 @flaky(max_runs=3)
-@pytest.mark.skipif(not PY3,
-                    reason="Only meant for Python 3")
 def test_multiprocessing(tmpdir):
     """
-    Test that multiprocessing works on Python 3.
+    Test that multiprocessing works.
     """
     # Command to start the kernel
     cmd = "from spyder_kernels.console import start; start.main()"
 
     with setup_kernel(cmd) as client:
         # Remove all variables
-        client.execute("%reset -f")
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive("%reset -f", timeout=TIMEOUT)
 
         # Write multiprocessing code to a file
         code = """
@@ -468,8 +545,8 @@ if __name__ == '__main__':
         p.write(code)
 
         # Run code
-        client.execute("runfile(r'{}')".format(to_text_string(p)))
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive(
+            "%runfile {}".format(repr(str(p))), timeout=TIMEOUT)
 
         # Verify that the `result` variable is defined
         client.inspect('result')
@@ -481,19 +558,16 @@ if __name__ == '__main__':
 
 
 @flaky(max_runs=3)
-@pytest.mark.skipif(not PY3,
-                    reason="Only meant for Python 3")
 def test_multiprocessing_2(tmpdir):
     """
-    Test that multiprocessing works on Python 3.
+    Test that multiprocessing works.
     """
     # Command to start the kernel
     cmd = "from spyder_kernels.console import start; start.main()"
 
     with setup_kernel(cmd) as client:
         # Remove all variables
-        client.execute("%reset -f")
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive("%reset -f", timeout=TIMEOUT)
 
         # Write multiprocessing code to a file
         code = """
@@ -515,8 +589,8 @@ if __name__ == '__main__':
         p.write(code)
 
         # Run code
-        client.execute("runfile(r'{}')".format(to_text_string(p)))
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive(
+            "%runfile {}".format(repr(str(p))), timeout=TIMEOUT)
 
         # Verify that the `result` variable is defined
         client.inspect('result')
@@ -529,19 +603,22 @@ if __name__ == '__main__':
 
 
 @flaky(max_runs=3)
-@pytest.mark.skipif(not PY3,
-                    reason="Only meant for Python 3")
+@pytest.mark.skipif(
+    sys.platform == 'darwin' and sys.version_info[:2] == (3, 8),
+    reason="Fails on Mac with Python 3.8")
+@pytest.mark.skipif(
+    os.environ.get('USE_CONDA') != 'true',
+    reason="Doesn't work with pip packages")
 def test_dask_multiprocessing(tmpdir):
     """
-    Test that dask multiprocessing works on Python 3.
+    Test that dask multiprocessing works.
     """
     # Command to start the kernel
     cmd = "from spyder_kernels.console import start; start.main()"
 
     with setup_kernel(cmd) as client:
         # Remove all variables
-        client.execute("%reset -f")
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive("%reset -f")
 
         # Write multiprocessing code to a file
         # Runs two times to verify that in the second case it doesn't break
@@ -557,11 +634,11 @@ if __name__=='__main__':
         p.write(code)
 
         # Run code two times
-        client.execute("runfile(r'{}')".format(to_text_string(p)))
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive(
+            "%runfile {}".format(repr(str(p))), timeout=TIMEOUT)
 
-        client.execute("runfile(r'{}')".format(to_text_string(p)))
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive(
+            "%runfile {}".format(repr(str(p))), timeout=TIMEOUT)
 
         # Verify that the `x` variable is defined
         client.inspect('x')
@@ -582,16 +659,15 @@ def test_runfile(tmpdir):
 
     with setup_kernel(cmd) as client:
         # Remove all variables
-        client.execute("%reset -f")
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive("%reset -f", timeout=TIMEOUT)
 
         # Write defined variable code to a file
-        code = u"result = 'hello world'; error # make an error"
+        code = "result = 'hello world'; error # make an error"
         d = tmpdir.join("defined-test.py")
         d.write(code)
 
         # Write undefined variable code to a file
-        code = dedent(u"""
+        code = dedent("""
         try:
             result3 = result
         except NameError:
@@ -601,9 +677,8 @@ def test_runfile(tmpdir):
         u.write(code)
 
         # Run code file `d` to define `result` even after an error
-        client.execute("runfile(r'{}', current_namespace=False)"
-                       .format(to_text_string(d)))
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive(
+            "%runfile {}".format(repr(str(d))), timeout=TIMEOUT)
 
         # Verify that `result` is defined in the current namespace
         client.inspect('result')
@@ -614,9 +689,8 @@ def test_runfile(tmpdir):
         assert content['found']
 
         # Run code file `u` without current namespace
-        client.execute("runfile(r'{}', current_namespace=False)"
-                       .format(to_text_string(u)))
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive(
+            "%runfile {}".format(repr(str(u))), timeout=TIMEOUT)
 
         # Verify that the variable `result2` is defined
         client.inspect('result2')
@@ -627,9 +701,8 @@ def test_runfile(tmpdir):
         assert content['found']
 
         # Run code file `u` with current namespace
-        client.execute("runfile(r'{}', current_namespace=True)"
-                       .format(to_text_string(u)))
-        msg = client.get_shell_msg(timeout=TIMEOUT)
+        msg = client.execute_interactive("%runfile {} --current-namespace"
+                                        .format(repr(str(u))), timeout=TIMEOUT)
         content = msg['content']
 
         # Verify that the variable `result3` is defined
@@ -650,6 +723,9 @@ def test_runfile(tmpdir):
 
 
 @flaky(max_runs=3)
+@pytest.mark.skipif(
+    sys.platform == 'darwin' and sys.version_info[:2] == (3, 8),
+    reason="Fails on Mac with Python 3.8")
 def test_np_threshold(kernel):
     """Test that setting Numpy threshold doesn't make the Variable Explorer slow."""
 
@@ -658,30 +734,27 @@ def test_np_threshold(kernel):
     with setup_kernel(cmd) as client:
 
         # Set Numpy threshold, suppress and formatter
-        client.execute("""
+        client.execute_interactive("""
 import numpy as np;
 np.set_printoptions(
     threshold=np.inf,
     suppress=True,
     formatter={'float_kind':'{:0.2f}'.format})
-    """)
-        client.get_shell_msg(timeout=TIMEOUT)
+    """, timeout=TIMEOUT)
 
         # Create a big Numpy array and an array to check decimal format
-        client.execute("""
+        client.execute_interactive("""
 x = np.random.rand(75000,5);
 a = np.array([123412341234.123412341234])
-""")
-        client.get_shell_msg(timeout=TIMEOUT)
+""", timeout=TIMEOUT)
 
         # Assert that NumPy threshold, suppress and formatter
         # are the same as the ones set by the user
-        client.execute("""
+        client.execute_interactive("""
 t = np.get_printoptions()['threshold'];
 s = np.get_printoptions()['suppress'];
 f = np.get_printoptions()['formatter']
-""")
-        client.get_shell_msg(timeout=TIMEOUT)
+""", timeout=TIMEOUT)
 
         # Check correct decimal format
         client.inspect('a')
@@ -716,21 +789,10 @@ f = np.get_printoptions()['formatter']
         assert "{'float_kind': <built-in method format of str object" in content
 
 
-# We declare this constant immediately before the test, as determining
-# that TURTLE_ACTIVE is True will briefly pop up a window, similar to the
-# windows that will pop up during the test itself.
-TURTLE_ACTIVE = False
-try:
-    import turtle
-    turtle.Screen()
-    turtle.bye()
-    TURTLE_ACTIVE = True
-except:
-    pass
-
 @flaky(max_runs=3)
-@pytest.mark.skipif(not TURTLE_ACTIVE,
-                    reason="Doesn't work on non-interactive settings or Python installations without Tk")
+@pytest.mark.skipif(
+    not TURTLE_ACTIVE,
+    reason="Doesn't work on non-interactive settings or Python without Tk")
 def test_turtle_launch(tmpdir):
     """Test turtle scripts running in the same kernel."""
     # Command to start the kernel
@@ -738,8 +800,7 @@ def test_turtle_launch(tmpdir):
 
     with setup_kernel(cmd) as client:
         # Remove all variables
-        client.execute("%reset -f")
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive("%reset -f", timeout=TIMEOUT)
 
         # Write turtle code to a file
         code = """
@@ -763,8 +824,8 @@ turtle.bye()
         p.write(code)
 
         # Run code
-        client.execute("runfile(r'{}')".format(to_text_string(p)))
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive(
+            "%runfile {}".format(repr(str(p))), timeout=TIMEOUT)
 
         # Verify that the `tess` variable is defined
         client.inspect('tess')
@@ -781,8 +842,8 @@ turtle.bye()
         p.write(code)
 
         # Run code again
-        client.execute("runfile(r'{}')".format(to_text_string(p)))
-        client.get_shell_msg(timeout=TIMEOUT)
+        client.execute_interactive(
+            "%runfile {}".format(repr(str(p))), timeout=TIMEOUT)
 
         # Verify that the `a` variable is defined
         client.inspect('a')
@@ -802,10 +863,8 @@ def test_matplotlib_inline(kernel):
     with setup_kernel(cmd) as client:
         # Get current backend
         code = "import matplotlib; backend = matplotlib.get_backend()"
-        client.execute(code, user_expressions={'output': 'backend'})
-        reply = client.get_shell_msg(timeout=TIMEOUT)
-        while 'user_expressions' not in reply['content']:
-            reply = client.get_shell_msg(timeout=TIMEOUT)
+        reply = client.execute_interactive(
+            code, user_expressions={'output': 'backend'}, timeout=TIMEOUT)
 
         # Transform value obtained through user_expressions
         user_expressions = reply['content']['user_expressions']
@@ -820,10 +879,7 @@ def test_do_complete(kernel):
     """
     Check do complete works in normal and debugging mode.
     """
-    if IPYKERNEL_6:
-        asyncio.run(kernel.do_execute('abba = 1', True))
-    else:
-        kernel.do_execute('abba = 1', True)
+    asyncio.run(kernel.do_execute('abba = 1', True))
     assert kernel.get_value('abba') == 1
     match = kernel.do_complete('ab', 2)
     assert 'abba' in match['matches']
@@ -832,7 +888,7 @@ def test_do_complete(kernel):
     pdb_obj = SpyderPdb()
     pdb_obj.curframe = inspect.currentframe()
     pdb_obj.completenames = lambda *ignore: ['baba']
-    kernel.shell.pdb_session = pdb_obj
+    kernel.shell._namespace_stack = [pdb_obj]
     match = kernel.do_complete('ba', 2)
     assert 'baba' in match['matches']
     pdb_obj.curframe = None
@@ -846,17 +902,11 @@ def test_callables_and_modules(kernel, exclude_callables_and_modules,
     Tests that callables and modules are in the namespace view only
     when the right options are passed to the kernel.
     """
-    if IPYKERNEL_6:
-        asyncio.run(kernel.do_execute('import numpy', True))
-        asyncio.run(kernel.do_execute('a = 10', True))
-        asyncio.run(kernel.do_execute('def f(x): return x', True))
-    else:
-        kernel.do_execute('import numpy', True)
-        kernel.do_execute('a = 10', True)
-        kernel.do_execute('def f(x): return x', True)
+    asyncio.run(kernel.do_execute('import numpy', True))
+    asyncio.run(kernel.do_execute('a = 10', True))
+    asyncio.run(kernel.do_execute('def f(x): return x', True))
 
     settings = kernel.namespace_view_settings
-
     settings['exclude_callables_and_modules'] = exclude_callables_and_modules
     settings['exclude_unsupported'] = exclude_unsupported
     nsview = kernel.get_namespace_view()
@@ -890,7 +940,7 @@ def test_comprehensions_with_locals_in_pdb(kernel):
     pdb_obj = SpyderPdb()
     pdb_obj.curframe = inspect.currentframe()
     pdb_obj.curframe_locals = pdb_obj.curframe.f_locals
-    kernel.shell.pdb_session = pdb_obj
+    kernel.shell._namespace_stack = [pdb_obj]
 
     # Create a local variable.
     kernel.shell.pdb_session.default('zz = 10')
@@ -916,7 +966,7 @@ def test_comprehensions_with_locals_in_pdb_2(kernel):
     pdb_obj = SpyderPdb()
     pdb_obj.curframe = inspect.currentframe()
     pdb_obj.curframe_locals = pdb_obj.curframe.f_locals
-    kernel.shell.pdb_session = pdb_obj
+    kernel.shell._namespace_stack = [pdb_obj]
 
     # Create a local variable.
     kernel.shell.pdb_session.default('aa = [1, 2]')
@@ -937,23 +987,17 @@ def test_namespaces_in_pdb(kernel):
     Test namespaces in pdb
     """
     # Define get_ipython for timeit
-    get_ipython = lambda: kernel.shell
+    get_ipython = lambda: kernel.shell  # noqa
     kernel.shell.user_ns["test"] = 0
     pdb_obj = SpyderPdb()
     pdb_obj.curframe = inspect.currentframe()
     pdb_obj.curframe_locals = pdb_obj.curframe.f_locals
-    kernel.shell.pdb_session = pdb_obj
+    kernel.shell._namespace_stack = [pdb_obj]
 
     # Check adding something to globals works
     pdb_obj.default("globals()['test2'] = 0")
     assert pdb_obj.curframe.f_globals["test2"] == 0
 
-    if PY2:
-        # no error method in py2
-        pdb_obj.curframe = None
-        pdb_obj.curframe_locals = None
-        return
-
     # Create wrapper to check for errors
     old_error = pdb_obj.error
     pdb_obj._error_occured = False
@@ -991,7 +1035,7 @@ def test_functions_with_locals_in_pdb(kernel):
     Frame = namedtuple("Frame", ["f_globals"])
     pdb_obj.curframe = Frame(f_globals=kernel.shell.user_ns)
     pdb_obj.curframe_locals = kernel.shell.user_ns
-    kernel.shell.pdb_session = pdb_obj
+    kernel.shell._namespace_stack = [pdb_obj]
 
     # Create a local function.
     kernel.shell.pdb_session.default(
@@ -1019,11 +1063,11 @@ def test_functions_with_locals_in_pdb_2(kernel):
 
     This is another regression test for spyder-ide/spyder-kernels#345
     """
-    baba = 1
+    baba = 1  # noqa
     pdb_obj = SpyderPdb()
     pdb_obj.curframe = inspect.currentframe()
     pdb_obj.curframe_locals = pdb_obj.curframe.f_locals
-    kernel.shell.pdb_session = pdb_obj
+    kernel.shell._namespace_stack = [pdb_obj]
 
     # Create a local function.
     kernel.shell.pdb_session.default(
@@ -1056,11 +1100,11 @@ def test_locals_globals_in_pdb(kernel):
     """
     Test thal locals and globals work properly in Pdb.
     """
-    a = 1
+    a = 1  # noqa
     pdb_obj = SpyderPdb()
     pdb_obj.curframe = inspect.currentframe()
     pdb_obj.curframe_locals = pdb_obj.curframe.f_locals
-    kernel.shell.pdb_session = pdb_obj
+    kernel.shell._namespace_stack = [pdb_obj]
 
     assert kernel.get_value('a') == 1
 
@@ -1095,52 +1139,96 @@ def test_locals_globals_in_pdb(kernel):
 
 
 @flaky(max_runs=3)
-@pytest.mark.parametrize("backend", [None, 'inline', 'tk', 'qt5'])
+@pytest.mark.parametrize("backend", [None, 'inline', 'tk', 'qt'])
 @pytest.mark.skipif(
-    not sys.platform.startswith('linux'),
-    reason="Doesn't work reliably on Windows and Mac")
-@pytest.mark.skipif(
-    not bool(os.environ.get('USE_CONDA')),
-    reason="Doesn't work with pip packages")
-@pytest.mark.skipif(
-    sys.version_info[:2] < (3, 8),
-    reason="Too flaky in Python 3.7 and doesn't work in older versions")
+    os.environ.get('USE_CONDA') != 'true',
+    reason="Doesn't work with pip packages"
+)
 def test_get_interactive_backend(backend):
     """
     Test that we correctly get the interactive backend set in the kernel.
     """
-    cmd = "from spyder_kernels.console import start; start.main()"
+    # This test passes locally but fails on CIs. Don't know why.
+    if sys.platform == "darwin" and backend == "qt" and os.environ.get('CI'):
+        return
 
+    cmd = "from spyder_kernels.console import start; start.main()"
     with setup_kernel(cmd) as client:
         # Set backend
         if backend is not None:
-            client.execute("%matplotlib {}".format(backend))
-            client.get_shell_msg(timeout=TIMEOUT)
-            client.execute("import time; time.sleep(.1)")
-            client.get_shell_msg(timeout=TIMEOUT)
+            client.execute_interactive(
+                "%matplotlib {}".format(backend), timeout=TIMEOUT
+            )
+            client.execute_interactive(
+                "import time; time.sleep(.1)", timeout=TIMEOUT
+            )
 
         # Get backend
         code = "backend = get_ipython().kernel.get_mpl_interactive_backend()"
-        client.execute(code, user_expressions={'output': 'backend'})
-        reply = client.get_shell_msg(timeout=TIMEOUT)
-        while 'user_expressions' not in reply['content']:
-            reply = client.get_shell_msg(timeout=TIMEOUT)
+        reply = client.execute_interactive(
+            code, user_expressions={'output': 'backend'}, timeout=TIMEOUT
+        )
 
         # Get value obtained through user_expressions
         user_expressions = reply['content']['user_expressions']
         value = user_expressions['output']['data']['text/plain']
 
+        # remove quotes
+        value = value[1:-1]
+
         # Assert we got the right interactive backend
         if backend is not None:
-            assert MPL_BACKENDS_FROM_SPYDER[value] == backend
+            assert value == backend
         else:
-            assert value == '0'
+            assert value == 'inline'
+
+
+def test_global_message(tmpdir):
+    """
+    Test that using `global` triggers a warning.
+    """
+    # Command to start the kernel
+    cmd = "from spyder_kernels.console import start; start.main()"
+
+    with setup_kernel(cmd) as client:
+        # Remove all variables
+        client.execute_interactive("%reset -f", timeout=TIMEOUT)
+
+        # Write code with a global to a file
+        code = (
+            "def foo1():\n"
+            "    global x\n"
+            "    x = 2\n"
+            "x = 1\n"
+            "print(x)\n"
+        )
+
+        p = tmpdir.join("test.py")
+        p.write(code)
+        global found
+        found = False
+
+        def check_found(msg):
+            if "text" in msg["content"]:
+                if ("WARNING: This file contains a global statement"  in
+                        msg["content"]["text"]):
+                    global found
+                    found = True
+
+        # Run code in current namespace
+        client.execute_interactive("%runfile {} --current-namespace".format(
+            repr(str(p))), timeout=TIMEOUT, output_hook=check_found)
+        assert not found
+
+        # Run code in empty namespace
+        client.execute_interactive(
+            "%runfile {}".format(repr(str(p))), timeout=TIMEOUT,
+            output_hook=check_found)
+
+        assert found
 
 
 @flaky(max_runs=3)
-@pytest.mark.skipif(
-    sys.version_info[0] < 3,
-    reason="Fails with python 2")
 def test_debug_namespace(tmpdir):
     """
     Test that the kernel uses the proper namespace while debugging.
@@ -1154,7 +1242,7 @@ def test_debug_namespace(tmpdir):
         d.write('def func():\n    bb = "hello"\n    breakpoint()\nfunc()')
 
         # Run code file `d`
-        msg_id = client.execute("runfile(r'{}')".format(to_text_string(d)))
+        client.execute("%runfile {}".format(repr(str(d))))
 
         # make sure that 'bb' returns 'hello'
         client.get_stdin_msg(timeout=TIMEOUT)
@@ -1181,6 +1269,103 @@ def test_debug_namespace(tmpdir):
                     break
 
 
+def test_interrupt():
+    """
+    Test that the kernel can be interrupted by calling a comm handler.
+    """
+    # Command to start the kernel
+    cmd = "from spyder_kernels.console import start; start.main()"
+    with setup_kernel(cmd) as client:
+        kernel_comm = CommBase()
+
+        # Create new comm and send the highest protocol
+        comm = Comm(kernel_comm._comm_name, client)
+        comm.open(data={})
+        comm._send_channel = client.control_channel
+        kernel_comm._register_comm(comm)
+
+        client.execute_interactive("import time", timeout=TIMEOUT)
+
+        # Try interrupting loop
+        t0 = time.time()
+        msg_id = client.execute("for i in range(100): time.sleep(.1)")
+        time.sleep(.2)
+        # Raise interrupt on control_channel
+        kernel_comm.remote_call().raise_interrupt_signal()
+        # Wait for shell message
+        while True:
+            assert time.time() - t0 < 5
+            msg = client.get_shell_msg(timeout=TIMEOUT)
+            if msg["parent_header"].get("msg_id") != msg_id:
+                # not from my request
+                continue
+            break
+        assert time.time() - t0 < 5
+
+        if os.name == 'nt':
+            # Windows doesn't do "interrupting sleep"
+            return
+
+        # Try interrupting sleep
+        t0 = time.time()
+        msg_id = client.execute("time.sleep(10)")
+        time.sleep(.2)
+        # Raise interrupt on control_channel
+        kernel_comm.remote_call().raise_interrupt_signal()
+        # Wait for shell message
+        while True:
+            assert time.time() - t0 < 5
+            msg = client.get_shell_msg(timeout=TIMEOUT)
+            if msg["parent_header"].get("msg_id") != msg_id:
+                # not from my request
+                continue
+            break
+        assert time.time() - t0 < 5
+
+
+def test_enter_debug_after_interruption():
+    """
+    Test that we can enter the debugger after interrupting the current
+    execution.
+    """
+    # Command to start the kernel
+    cmd = "from spyder_kernels.console import start; start.main()"
+    with setup_kernel(cmd) as client:
+        kernel_comm = CommBase()
+
+        # Create new comm and send the highest protocol
+        comm = Comm(kernel_comm._comm_name, client)
+        comm.open(data={})
+        comm._send_channel = client.control_channel
+        kernel_comm._register_comm(comm)
+
+        client.execute_interactive("import time", timeout=TIMEOUT)
+
+        # Try interrupting loop
+        t0 = time.time()
+        msg_id = client.execute("for i in range(100): time.sleep(.1)")
+        time.sleep(.2)
+        # Request to enter the debugger
+        kernel_comm.remote_call().request_pdb_stop()
+        # Wait for debug message
+        while True:
+            assert time.time() - t0 < 5
+            msg = client.get_iopub_msg(timeout=TIMEOUT)
+            if msg.get('msg_type') == 'stream':
+                print(msg["content"].get("text"))
+            if msg["parent_header"].get("msg_id") != msg_id:
+                # not from my request
+                continue
+            if msg.get('msg_type') == 'comm_msg':
+                if msg["content"].get("data", {}).get("content", {}).get(
+                        'call_name') == 'pdb_input':
+                    # pdb entered
+                    break
+                comm.handle_msg(msg)
+
+        assert time.time() - t0 < 5
+
+
 def test_non_strings_in_locals(kernel):
     """
     Test that we can hande non-string entries in `locals` when bulding the
@@ -1188,30 +1373,65 @@ def test_non_strings_in_locals(kernel):
 
     This is a regression test for issue spyder-ide/spyder#19145
     """
-    if IPYKERNEL_6:
-        execute = asyncio.run(kernel.do_execute(
-            'locals().update({1:2})', True))
-    else:
-        execute = kernel.do_execute('locals().update({1:2})', True)
-
+    asyncio.run(kernel.do_execute('locals().update({1:2})', True))
     nsview = repr(kernel.get_namespace_view())
     assert "1:" in nsview
 
 
-@pytest.mark.skipif(
-    sys.version_info[0] < 3, reason="Doesn't work with Python 2")
 def test_django_settings(kernel):
     """
     Test that we don't generate errors when importing `django.conf.settings`.
 
     This is a regression test for issue spyder-ide/spyder#19516
     """
-    execute = asyncio.run(kernel.do_execute(
-        'from django.conf import settings', True))
-
+    asyncio.run(kernel.do_execute('from django.conf import settings', True))
     nsview = repr(kernel.get_namespace_view())
     assert "'settings':" in nsview
 
 
+def test_hard_link_pdb(tmpdir):
+    """
+    Test that breakpoints on a file are recognised even when the path is
+    different.
+    """
+    # Create a file and a hard link
+    d = tmpdir.join("file.py")
+    d.write('def func():\n    bb = "hello"\n')
+    folder = tmpdir.join("folder")
+    os.mkdir(folder)
+    hard_link = folder.join("file.py")
+    os.link(d, hard_link)
+
+    # Make sure both paths point to the same file
+    assert os.path.samefile(d, hard_link)
+
+    # Make sure canonic returns the same path for a single file
+    pdb_obj = SpyderPdb()
+    assert pdb_obj.canonic(str(d)) == pdb_obj.canonic(str(hard_link))
+
+
+@pytest.mark.skipif(not os.environ.get('CI'), reason="Only works on CIs")
+def test_get_pythonenv_info(kernel):
+    """Test the output we get from this method."""
+    output = kernel.get_pythonenv_info()
+    assert output["path"] == sys.executable
+
+    if os.environ.get('USE_CONDA'):
+        assert output["name"] == "test"
+        assert output["env_type"] == PythonEnvType.Conda
+    else:
+        assert output["env_type"] in [
+            # This Custom here accounts for Linux packagers that run our tests
+            # in their CIs
+            PythonEnvType.Custom,
+            PythonEnvType.Conda,
+        ]
+
+    # Check these keys are present. Otherwise we'll break Spyder.
+    assert output["python_version"] == sys.version.split()[0]
+    assert output["ipython_version"] == ipython_release.version
+    assert output["sys_version"] == sys.version
+
+
 if __name__ == "__main__":
     pytest.main()
